@@ -3,16 +3,49 @@
 # ###
 from __future__ import annotations
 import warnings
-from typing import List
+from typing import List, Tuple
 import itertools
 from .exceptions import MolecularSpeciesError
 
+
+def flatten(l):
+    return [item for sublist in l for item in sublist]
+
+
+class DBPos:
+
+    __slots__ = ["loc", "geom"]
+
+    def __init__(self, loc: int, geom: str):
+        self.loc = loc
+        self.geom = geom.upper()
+
+    def __str__(self):
+        return str(self.loc) + self.geom
+
+    def __lt__(self, other):
+        if self.loc < other.loc:
+            return True
+        elif self.loc == other.loc:
+            return self.geom < other.geom
+        else:
+            return False
+
+    def __repr__(self):
+        return self.__str__()
+
+    def __eq__(self, other):
+        return self.loc == other.loc
+
+
 class FA:
 
-    __slots__ = ["__C", "__DB", "__OH", "__modifications", "__is_long_chain_base", "__is_ether_bond", "__ether_type"]
+    __slots__ = ["__C", "__DB", "__OH", "__modifications", "__is_long_chain_base", "__is_ether_bond",
+                 "__ether_type", "__db_positions"]
 
     def __init__(self, c: int, db: int, oh: int = None, modifications: list = None,
-                 is_long_chain_base: bool = False, is_ether_bond: bool = False, ether_type: str = None):
+                 is_long_chain_base: bool = False, is_ether_bond: bool = False, ether_type: str = None,
+                 db_positions: List[DBPos] = None):
         """
         Fatty acid class. Describes a fatty acid that can be part of a Lipid
         :param c: int, Number of carbons
@@ -43,9 +76,19 @@ class FA:
         if self.__is_long_chain_base and self.__is_ether_bond:
             warnings.warn("Fatty acid is long chain base and ether bond at once.")
 
+        if db_positions is None:
+            self.__db_positions = None
+        else:
+            if len(db_positions) != db:
+                warnings.warn("DB positions does not match the number of double bonds")
+            self.__db_positions = sorted(db_positions)
+
     def length(self) -> int:
         return self.__C
 
+    def db_positions(self) -> List[DBPos]:
+        return self.__db_positions
+
     def db_index(self) -> int:
         return self.__DB
 
@@ -64,13 +107,20 @@ class FA:
     def get_ether_type(self) -> str:
         return self.__ether_type
 
+    def _db_pos_equal(self, other: FA) -> bool:
+        if self.db_positions() is None or other.db_positions() is None:
+            return True
+        else:
+            return self.__db_positions == other.db_positions()
+
     def __eq__(self, other) -> bool:
         if not isinstance(other, FA):
             raise TypeError("Only comparisons between FAs allowed.")
         return (self.__C == other.length()) & (self.__DB == other.db_index()) &\
                (self.__OH == other.hydroxylations()) & (self.__modifications == other.modifications()) &\
                (self.__is_long_chain_base == other.is_long_chain_base()) &\
-               (self.__is_ether_bond == other.is_ether_bond())
+               (self.__is_ether_bond == other.is_ether_bond()) &\
+            self._db_pos_equal(other)
 
     def __str__(self) -> str:
         out_str = ""
@@ -78,7 +128,15 @@ class FA:
             out_str += self.__ether_type
         # if self.__is_long_chain_base:
         #     out_str += "d"
+        if self.modifications() is not None:
+            if len(self.modifications()) > 0:
+                for mod in self.modifications():
+                    out_str += str(mod + '-')
         out_str += f"{self.__C}:{self.__DB}"
+        if self.__db_positions is not None:
+            out_str += "("
+            out_str += ",".join([str(x) for x in self.db_positions()])
+            out_str += ")"
         if self.__OH:
             if self.__OH > 1:
                 out_str += f";O{self.__OH}"
@@ -146,17 +204,31 @@ class FA:
                     elif myoh > otheroh:
                         return False
 
-    def modified_fa(self, c, db, oh) -> FA:
-        return FA(self.__C+c, self.__DB+db, self.__OH+oh, self.__modifications, self.__is_long_chain_base,
-                  self.__is_ether_bond, self.__ether_type)
+    def modified_fa(self, c, db, oh, added_db_positions: List[DBPos] = None) -> FA:
+        """
+        Adding increasing length and unsaturation -> The other way will introduce problems with DB positions
+
+        """
+        if db == 0:
+            return FA(self.__C + c, self.__DB + db, self.__OH + oh, self.__modifications, self.__is_long_chain_base,
+                      self.__is_ether_bond, self.__ether_type, db_positions=self.__db_positions)
+        elif db > 0 and added_db_positions is None:
+            return FA(self.__C + c, self.__DB + db, self.__OH + oh, self.__modifications, self.__is_long_chain_base,
+                      self.__is_ether_bond, self.__ether_type, db_positions=None)
+        else:
+            new_db_positions = sorted(list(set(self.__db_positions + added_db_positions)))
+            if len(new_db_positions) < (len(self.__db_positions) + len(added_db_positions)):
+                warnings.warn("Duplicated Double bond positions added")
+            return FA(self.__C + c, self.__DB + db, self.__OH + oh, self.__modifications, self.__is_long_chain_base,
+                      self.__is_ether_bond, self.__ether_type, db_positions=new_db_positions)
 
     def as_ether(self, ether_type: str) -> FA:
         return FA(self.__C, self.__DB, self.__OH, self.__modifications, self.__is_long_chain_base,
-                  is_ether_bond=True, ether_type=ether_type)
+                  is_ether_bond=True, ether_type=ether_type, db_positions=self.__db_positions)
 
     def as_not_ether(self) -> FA:
         return FA(self.__C, self.__DB, self.__OH, self.__modifications, self.__is_long_chain_base,
-                  is_ether_bond=False)
+                  is_ether_bond=False, db_positions=self.__db_positions)
 
 
 class Lipid:
@@ -278,6 +350,12 @@ class Lipid:
         else:
             return self.__sum_ohs
 
+    def sum_fa_modifications(self) -> List[str]:
+        if self._is_molecular_species:
+            return flatten([x.modifications() for x in self.__fatty_acids if x.modifications() is not None])
+        else:
+            return self.__sum_fa_modifications
+
     def get_non_fa_modifications(self) -> List:
         return self.__non_fa_modifications
 
@@ -431,6 +509,10 @@ class Lipid:
             out_str = f"{clsstr}("
         else:
             out_str = f"{self.__lipid_class}("
+        if self.sum_fa_modifications() is not None:
+            if len(self.sum_fa_modifications()) > 0:
+                for mod in self.sum_fa_modifications():
+                    out_str += str(mod + '-')
         if self.get_ether():
             out_str += self.get_ether_type()
         # if self.get_number_long_chain_base():
@@ -452,6 +534,9 @@ class Lipid:
     def converted_to_mol_spec(self) -> bool:
         return self.__converted_to_mol_spec
 
+    def set_converted_to_mol_spec(self, converted=True):
+        self.__converted_to_mol_spec = converted
+
     def get_native_string(self) -> str:
         if self.converted_to_mol_spec():
             return self.sum_species_str()
@@ -464,6 +549,7 @@ class Lipid:
     def get_div_factor(self):
         return self.__div_factor
 
+
 def molecular_equals_sum_species(ms: Lipid, ss: Lipid) -> bool:
     if ms.is_molecular_species() and not ss.is_molecular_species():
         if ms.get_ether() or ss.get_ether():
@@ -491,6 +577,7 @@ def molecular_equals_sum_species(ms: Lipid, ss: Lipid) -> bool:
         warnings.warn("Comparison requires a molecular and a sum species lipid.")
         return False
 
+
 if __name__ == "__main__":
     fa1 = FA(20, 4, None)
     fa2 = FA(18, 0, 0)
